import os
import ssl
import logging
from six.moves.urllib_parse import urlparse
from ._common import BaseServer
from ._common import ServerConnection
from ._service import Service
from arcgis.gis import GIS

_log = logging.getLogger()
########################################################################
class ServicesDirectory(BaseServer):
    """
    Provides ArcGIS Server Services Directory access.

    The Services Directory allows you to browse the contents of an ArcGIS Server and obtain
    information that can be useful to endusers for developing applications, performing analysis, or
    cataloging services. The ServicesDirectory is a view of the ArcGIS Server REST API in a python
    class.

    With the Services Directory, you can do the following:
     - Browse the contents of the GIS Server and get service-level metadata
     - Navigate a series of links to view information about the services on your GIS Server.
     - Get information to help you develop applications

    When you develop applications with the Python API, you must provide URLs to services and the
    layers and functionality they expose. The Services Directory provides an interactive way for
    you to construct those URLs.

    The Services Directory works using REST. REST is an architectural style that allows ArcGIS
    Server to reveal a hierarchy of information about itself through endpoints, or URLs. When you
    use the Services Directory, you navigate through a series of links to discover information
    about the server. Each time you click a link, you see a new page that reveals additional
    information about what's available on the server. The information that you see on the page is
    retrieved through REST using the page's URL.


    =====================     ====================================================================
    **Arguments**             **Description**
    ---------------------     --------------------------------------------------------------------
    url                       string required. The web address to the ArcGIS Server administration
                              end point.

                              Example: https://mysite.com/arcgis

                              The URL should be formatted as follows:
                              <scheme>://<host>:<port (optional)>/<web adapter>
    ---------------------     --------------------------------------------------------------------
    baseurl                   optional string, the root URL to a site.
                              Example: https://mysite.com/arcgis
    ---------------------     --------------------------------------------------------------------
    tokenurl                  optional string. Used when a site if federated or when the token
                              URL differs from the site's baseurl.  If a site is federated, the
                              token URL will return as the Portal token and ArcGIS Server users
                              will not validate correctly.
    ---------------------     --------------------------------------------------------------------
    username                  optional string, login username for BUILT-IN security
    ---------------------     --------------------------------------------------------------------
    password                  optional string, a secret word or phrase that must be used to gain
                              access to the account above.
    ---------------------     --------------------------------------------------------------------
    key_file                  optional string, path to PKI ket file
    ---------------------     --------------------------------------------------------------------
    cert_file                 optional string, path to PKI cert file
    ---------------------     --------------------------------------------------------------------
    proxy_host                optional string, web address to the proxy host

                              Example: proxy.mysite.com
    ---------------------     --------------------------------------------------------------------
    proxy_port                optional integer, default is 80. The port where the proxy resided on
    ---------------------     --------------------------------------------------------------------
    expiration                optional integer. The Default is 60. This is the length of time a
                              token is valid for.
                              Example 1440 is one week.
    ---------------------     --------------------------------------------------------------------
    all_ssl                   optional boolean. The default is False. If True, all calls will be
                              made over HTTPS instead of HTTP.
    ---------------------     --------------------------------------------------------------------
    portal_connection         optional GIS. This is used when a site is federated. It is the
                              ArcGIS Online or Portal GIS object used.
    ---------------------     --------------------------------------------------------------------
    initialize                optional boolean.  The default is False.  If True, the object will
                              attempt to reach out to the URL resource and populate at creation
                              time.
    ---------------------     --------------------------------------------------------------------
    verify_cert               optional booean.  The defaut is True.  If False, any certificate
                              errors will be ignored.  This is useful for new sites or where site
                              that have invalid certificates.
    =====================     ====================================================================

    """
    _con = None
    _gis = None
    _url = None
    _adminurl = None
    _properties = None
    #----------------------------------------------------------------------
    def __init__(self, url,
                 username=None,
                 password=None,
                 key_file=None,
                 cert_file=None,
                 verify_cert=False,
                 **kwargs):
        """Constructor"""
        super(ServicesDirectory, self)

        if url.lower().find('/rest') == -1 and \
           url.endswith('/rest') == False:
            url = "%s/rest/services" % url
        if url.lower().find('/services') == -1 and \
           url.lower().endswith('/services') == False:
            url = "%s/services" % url
        self._url = url
        self._username = username
        self._password = password
        self._key_file = key_file
        self._cert_file = cert_file
        self._portal_connection = kwargs.pop('portal_connection', None)
        self._is_agol = kwargs.pop("is_agol", False)
        con = kwargs.pop('con', None)
        if verify_cert == False:
            import ssl
            ssl._create_default_https_context = ssl._create_unverified_context
        aurl = None
        if 'admin_url' in kwargs:
            aurl = kwargs.pop('admin_url', None)
        if aurl is None:
            parsed = urlparse(url)
            wa = parsed.path[1:].split('/')[0]
            self._adminurl = "%s://%s/%s/admin" % (parsed.scheme, parsed.netloc, wa)
        else:
            self._adminurl = aurl

        if self._is_agol and self._portal_connection:
            if isinstance(self._portal_connection, GIS):
                self._con = self._portal_connection._portal.con
            elif hasattr(self._portal_connection, 'post'):
                self._con = self._portal_connection
        elif con:
            self._con = con
        else:
            self._con = ServerConnection(baseurl=url,
                                         username=username,
                                         password=password,
                                         key_file=key_file,
                                         cert_file=cert_file,
                                         portal_connection=self._portal_connection,
                                         verfiy_cert=verify_cert,
                                         **kwargs)
        self._gis = kwargs.pop('gis', None)
        if self._is_agol == False and self._con._auth.lower() != "anon":
            try:
                from .admin.administration import Server
                self.admin = Server(gis=self._con,
                                    url=self._adminurl,
                                    servicesdirectory=self,
                                    initialize=False)
            except: pass
        self._init(self._con)
    #----------------------------------------------------------------------
    def __str__(self):
        return '<%s at %s>' % (type(self).__name__, self.url)
    #----------------------------------------------------------------------
    def __repr__(self):
        return '<%s at %s>' % (type(self).__name__, self.url)
    #----------------------------------------------------------------------
    def report(self, as_html=True, folder=None):
        """
        Generates a table of Services in the given folder, as a Pandas dataframe.


        """
        import pandas as pd
        pd.set_option('display.max_colwidth', -1)
        data = []
        a_template = """<a href="%s?token=%s">URL Link</a>"""
        columns = ['Service Name', 'Service URL']
        if folder is None:
            res = self._con.get(self._url, {"f" : 'json'})
        elif folder.lower() in [f.lower() for f in self.folders]:
            res = self._con.get("%s/%s" % (self._url, folder), {"f" : 'json'})
        if 'services' in res:
            for s in res['services']:
                #if s['name'].split('/')[-1].lower() == name.lower():
                url = "%s/%s/%s" % (self._url,
                                  s['name'],
                                  s['type'])
                data.append([s['name'].split('/')[-1], """<a href="%s">Service</a>""" % url])
        #for service in self.list(folder=folder):
            #name = os.path.basename(os.path.dirname(service._url))
            #data.append([name, a_template % (service._url, self._con.token)])
            #del service
        df = pd.DataFrame(data=data, columns=columns)
        if as_html:
            table = \
                """<div class="9item_container" style="height: auto; overflow: hidden; """ + \
                """border: 1px solid #cfcfcf; border-radius: 2px; background: #f6fafa; """ + \
                """line-height: 1.21429em; padding: 10px;">%s</div>""" % df.to_html(escape=False,
                                                                                    index=False)
            return table.replace('\n', '')
        else:
            return df

    #----------------------------------------------------------------------
    def get(self, name, folder=None):
        """returns a single service in a folder"""
        if folder is None:
            res = self._con.get(self._url, {"f" : 'json'})
        elif folder.lower() in [f.lower() for f in self.folders]:
            res = self._con.get("%s/%s" % (self._url, folder), {"f" : 'json'})
        if 'services' in res:
            for s in res['services']:
                if s['name'].split('/')[-1].lower() == name.lower():
                    return Service(url="%s/%s/%s" % (self._url,
                                                     s['name'],
                                                     s['type']),
                                   server=self._con)
                del s
        return None
    #----------------------------------------------------------------------
    def list(self, folder=None):
        """
        returns a list of services at the given folder
        """
        services = []
        if folder is None:
            res = self._con.get(self._url, {"f" : 'json'})
        elif folder.lower() in [f.lower() for f in self.folders]:
            res = self._con.get("%s/%s" % (self._url, folder), {"f" : 'json'})
        if 'services' in res:
            for s in res['services']:
                try:
                    services.append(Service(url="%s/%s/%s" % (self._url,
                                                              s['name'],
                                                              s['type']),
                                            server=self._con))

                except:
                    url ="%s/%s/%s" % (self._url, s['name'], s['type'])
                    _log.warning("Could not load service: %s" % url)
        return services
    #----------------------------------------------------------------------
    def find(self, service_name, folder=None):
        """
        finds a service based on it's name in a given folder
        """
        return self.get(name=service_name, folder=folder)
    #----------------------------------------------------------------------
    @property
    def folders(self):
        """
        returns a list of server folders
        """
        self._init()
        if self._is_agol:
            return ['/']
        else:
            return self.properties['folders']
        return []


"""
These functions are used for both the day-to-day management of geographic data and for combining data prior to analysis.

dissolve_boundaries merges together areas that share a common boundary and a common attribute value.
extract_data creates new datasets by extracting features from your existing data.
merge_layers copies all the features from two or more existing layers into a new layer.
overlay_layers combines two or more layers into one single layer. You can think of overlay as peering through a stack of
maps and creating a single map containing all the information found in the stack.
"""
import arcgis as _arcgis

def dissolve_boundaries(
        input_layer,
        dissolve_fields=[],
        summary_fields=[],
        output_name=None,
        context=None,
        gis=None):
    """
    Dissolve features based on specified fields.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The layer containing polygon features that will be dissolved.
    dissolve_fields : Optional list of strings
        One or more fields from the input that control which polygons are merged. If no fields are supplied, all
        polygons that overlap or shared a common border will be dissolved into one polygon.
    summary_fields : Optional list of strings
        A list of field names and statistical types that will be used to summarize the output. Supported statistics
        include: Sum, Mean, Min, Max, and Stddev.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    dissolved_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.dissolve_boundaries(
        input_layer,
        dissolve_fields,
        summary_fields,
        output_name,
        context)


def extract_data(
        input_layers,
        extent=None,
        clip=False,
        data_format=None,
        output_name=None,
        context=None,
        gis=None):
    """
    Select and download data for a specified area of interest. Layers that you select will be added to a zip file or
    layer package.

    Parameters
    ----------
    input_layers : Required list of strings
        The layers from which you can extract features.
    extent : Optional string
        The area that defines which features will be included in the output zip file or layer package.
    clip : Optional bool
        Select features that intersect the extent or clip features within the extent.
    data_format : Optional string
        Format of the data that will be extracted and downloaded.  Layer packages will always include file geodatabases.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    content_id : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.extract_data(
        input_layers,
        extent,
        clip,
        data_format,
        output_name,
        context)


def merge_layers(
        input_layer,
        merge_layer,
        merging_attributes=[],
        output_name=None,
        context=None,
        gis=None):
    """
    Combines two inputs of the same feature data type into a new output.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
         The point, line, or polygon  features to merge with the mergeLayer.
    merge_layer : Required layer (see Feature Input in documentation)
        The point, line or polygon features to merge with inputLayer.  mergeLayer must contain the same feature type
        point, line, or polygon) as the inputLayer.
    merging_attributes : Optional list of strings
        An array of values that describe how fields from the mergeLayer are to be modified.  By default all fields from
        both inputs will be carried across to the output.
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    merged_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.merge_layers(
        input_layer,
        merge_layer,
        merging_attributes,
        output_name,
        context)


def overlay_layers(
        input_layer,
        overlay_layer,
        overlay_type="Intersect",
        snap_to_input=False,
        output_type="Input",
        tolerance=None,
        output_name=None,
        context=None,
        gis=None):
    """
    Overlays the input layer with the overlay layer. Overlay operations supported are Intersect, Union, and Erase.

    Parameters
    ----------
    input_layer : Required layer (see Feature Input in documentation)
        The input analysis layer.
    overlay_layer : Required layer (see Feature Input in documentation)
        The layer to be overlaid with the analysis layer.
    overlay_type : Optional string
        The overlay type (INTERSECT, UNION, or ERASE) defines how the analysis layer and the overlay layer are combined.
    snap_to_input : Optional bool
        When the distance between features is less than the tolerance, the features in the overlay layer will snap to
        the features in the input layer.
    output_type : Optional string
        The type of intersection (INPUT, LINE, POINT).
    tolerance : Optional float
        The minimum distance separating all feature coordinates (nodes and vertices) as well as the distance a
        coordinate can move in X or Y (or both).
    output_name : Optional string
        Additional properties such as output feature service name.
    context : Optional string
        Additional settings such as processing extent and output spatial reference.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    output_layer : layer (FeatureCollection)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    return gis._tools.featureanalysis.overlay_layers(
        input_layer,
        overlay_layer,
        overlay_type,
        snap_to_input,
        output_type,
        tolerance,
        output_name,
        context)

def create_route_layers(
        route_data_item,
        delete_route_data_item=False,
        tags=None,
        summary=None,
        route_name_prefix=None,
        folder_name=None,
        gis=None):
    """
    Creates route layer items on the portal from the input route data. 

    Parameters
    ----------
    route_data_item : Required item
        The route data item that is used to create route layer items.
    delete_route_data_item : Required Boolean (see Feature Input in documentation)
        Indicates if the input route data item should be deleted. The default value is False which does not delete the route data item.
    tags: Optional string
        Tags used to describe and identify the route layer items. Individual tags are separated using a comma. The route name is always
        added as a tag even when a value for this argument is not specified. 
    summary: Optional string
        The summary displayed as part of the item information for the route layer item. If a value for this argument is not specified,
        a default summary text "Route and directions for <Route Name>" is used.
    route_name_prefix : Optional string
        A qualifier added to the title of every route layer item. This can be used to designate all routes that are shared for a 
        specific purpose to have the same prefix in the title. The name of the route is always appended after this qualifier.
        If a value for the route_name_prefix is not specified, the title for the route layer item is created using only the route name.
    folder_name: Optional string
        The folder within your personal online workspace (My Content in your ArcGIS Online or Portal for ArcGIS organization) where the
        route layer items will be created. If a folder with the specified name does not exist, a new folder will be created.
        If a folder with the specified name exists, the items will be created in the existing folder.
        If a value for folder_name is not specified, the route layer items are created in the root folder of your online workspace.
    gis :
        Optional, the GIS on which this tool runs. If not specified, the active GIS is used.

    Returns
    -------
    route_layers : list (items)
    """
    gis = _arcgis.env.active_gis if gis is None else gis
    output_name = {}
    output_item_properties = {}
    if route_name_prefix:
        output_item_properties["title"] = route_name_prefix
    if tags:
        output_item_properties["tags"] = tags
    if summary:
        output_item_properties["snippet"] = summary
    if folder_name:
        folder_id = ""
        # Get a dict of folder names for the current user
        folders = {fld["title"]: fld for fld in gis.users.me.folders}
        # if the folder already exists, just get its folder id
        if folder_name in folders:
            folder_id = folders[folder_name].get("id", "")
        else:
            # Create a new folder and get its folder id
            new_folder = gis.content.create_folder(folder_name)
            folder_id = new_folder.get("id", "")
        if folder_id:
            output_item_properties["folderId"] = folder_id
    if output_item_properties:
        output_name["itemProperties"] = output_item_properties

    return gis._tools.featureanalysis.create_route_layers(
        route_data_item,
        delete_route_data_item,
        output_name)



#!/usr/bin/env python

# Thanks @takluyver for your cite2c install.py.
# Copyright (c) IPython Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function

import argparse
from os.path import dirname, abspath, join as pjoin

def install(user=False, symlink=False, enable=False):
    """Install the widget nbextension and optionally enable it.

    Parameters
    ----------
    user: bool
        Install for current user instead of system-wide.
    symlink: bool
        Symlink instead of copy (for development).
    """
    try:
        from notebook.nbextensions import install_nbextension
        from notebook.services.config import ConfigManager
    except ModuleNotFoundError:
        print('"notebook" not installed, silently failing...')
        return
    widgetsdir = pjoin(dirname(abspath(__file__)), 'widgets')
    install_nbextension(widgetsdir, destination='arcgis', user=user, symlink=symlink)

    cm = ConfigManager()
    cm.update('notebook', {
        "load_extensions": {
            "arcgis/mapview": True,
        }
    })

def uninstall():
    try:
        from notebook.nbextensions import uninstall_nbextension

        """Unnstall the widget nbextension from user and system locations
        """
        print("Uninstalling prior versions of arcgis widget")
        uninstall_nbextension('arcgis', user=True)
        uninstall_nbextension('arcgis', user=False)
    except:
        print('Manually uninstall any prior version of arcgis widget using:\n\t"jupyter nbextension uninstall arcgis --user" and \n\t"jupyter nbextension uninstall arcgis"')

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description="Installs the ArcGIS IPython widgets")
    parser.add_argument("-u", "--user", help="Install as current user instead of system-wide", action="store_true")
    parser.add_argument("-s", "--symlink", help="Symlink instead of copying files", action="store_true")
    parser.add_argument("-r", "--remove", help="Remove i.e. uninstall the extension", action="store_true")
    args = parser.parse_args()

    if args.remove:
        uninstall()
    else:
        install(user=args.user, symlink=args.symlink)



"""
Reads shapefiles, feature classes, table into a spatial dataframe
"""
from __future__ import print_function
from __future__ import division
import os
import six
import copy
import logging
import tempfile
from warnings import warn
import numpy as np
import pandas as pd
from six import iteritems, integer_types
from datetime import datetime
from six.moves.urllib import request
from ..utils import NUMERIC_TYPES, STRING_TYPES, DATETIME_TYPES
from ..utils import sanitize_field_name
from arcgis.geometry import _types
try:
    import arcpy
    from arcpy import da
    HASARCPY = True
except:
    HASARCPY = False
try:
    import shapefile
    HASPYSHP = True
except:
    HASPYSHP = False
try:
    import fiona
    HASFIONA = True
except:
    HASFIONA = False

_log=logging.getLogger(__name__)

def _from_xy(df, x_column, y_column, sr=None):
    """

    """
    from arcgis.geometry import SpatialReference, Geometry
    from arcgis.features import SpatialDataFrame
    if sr is None:
        sr = SpatialReference({'wkid' : 4326})
    if not isinstance(sr, SpatialReference):
        if isinstance(sr, dict):
            sr = SpatialReference(sr)
        elif isinstance(sr, int):
            sr = SpatialReference({'wkid' : sr})
        elif isinstance(sr, str):
            sr = SpatialReference({'wkt' : sr})
    geoms = []
    for idx, row in df.iterrows():
        geoms.append(
            Geometry({'x' : row[x_column], 'y' : row[y_column],
             'spatialReference' : sr})
        )
    df['SHAPE'] = geoms
    return SpatialDataFrame(data=df, sr=sr)

def _pyshp_to_shapefile(df, out_path, out_name):
    """
    Saves a SpatialDataFrame to a Shapefile using pyshp

    :Parameters:
     :df: spatial dataframe
     :out_path: folder location to save the data
     :out_name: name of the shapefile
    :Output:
     path to the shapefile or None if pyshp isn't installed or
     spatial dataframe does not have a geometry column.
    """
    from .....geometry._types import Geometry
    if HASPYSHP:
        GEOMTYPELOOKUP = {
            "Polygon" : shapefile.POLYGON,
            "Point" : shapefile.POINT,
            "Polyline" : shapefile.POLYLINE,
            'null' : shapefile.NULL
        }
        if os.path.isdir(out_path) == False:
            os.makedirs(out_path)
        out_fc = os.path.join(out_path, out_name)
        if out_fc.lower().endswith('.shp') == False:
            out_fc += ".shp"
        geom_field = df.geometry.name
        if geom_field is None:
            return
        geom_type = "null"
        idx = df[geom_field].first_valid_index()
        if idx > -1:
            geom_type = df.loc[idx][geom_field].type
        shpfile = shapefile.Writer(GEOMTYPELOOKUP[geom_type])
        shpfile.autoBalance = 1
        for c in df.columns:
            idx = df[c].first_valid_index()
            if idx > -1:
                if isinstance(df[c].loc[idx],
                              Geometry):
                    geom_field = (c, "GEOMETRY")
                else:
                    if isinstance(df[c].loc[idx], six.string_types):
                        shpfile.field(name=c, size=255)
                    elif isinstance(df[c].loc[idx], six.integer_types):
                        shpfile.field(name=c, fieldType="N", size=5)
                    elif isinstance(df[c].loc[idx], (np.int, np.int32, np.int64)):
                        shpfile.field(name=c, fieldType="N", size=10)
                    elif isinstance(df[c].loc[idx], (np.float, np.float64)):
                        shpfile.field(name=c, fieldType="F", size=19, decimal=11)
                    elif isinstance(df[c].loc[idx], (datetime, np.datetime64)):
                        shpfile.field(name=c, fieldType="D", size=8)
                    elif isinstance(df[c].loc[idx], (bool, np.bool)):
                        shpfile.field(name=c, fieldType="L", size=1)
            del c
            del idx
        for idx, row in df.iterrows():
            geom = row[df.geometry.name]
            if geom.type == "Polygon":
                shpfile.poly(geom['rings'])
            elif geom.type == "Polyline":
                shpfile.line(geom['paths'])
            elif geom.type == "Point":
                shpfile.point(x=geom.x, y=geom.y)
            else:
                shpfile.null()
            shpfile.record(*row.tolist()[:-1])
            del idx
            del row
            del geom
        shpfile.save(out_fc)


        # create the PRJ file
        try:
            wkid = df.sr['wkid']
            try:
                wkid = df.sr['latestWkid']
            except:
                pass # try and use wkid instead

            prj_filename = out_fc.replace('.shp', '.prj')

            url = 'http://epsg.io/{}.esriwkt'.format(wkid)

            opener = request.build_opener()
            opener.addheaders = [('User-Agent', 'geosaurus')]
            resp = opener.open(url)

            wkt = resp.read().decode('utf-8')
            if len(wkt) > 0:
                prj = open(prj_filename, "w")
                prj.write(wkt)
                prj.close()
        except:
            # Unable to write PRJ file.
            pass

        del shpfile
        return out_fc
    return None
#--------------------------------------------------------------------------
def from_featureclass(filename, **kwargs):
    """
    Returns a GeoDataFrame from a feature class.
    Inputs:
     filename: full path to the feature class
    Optional Parameters:
     sql_clause: sql clause to parse data down
     where_clause: where statement
     sr: spatial reference object
     fields: list of fields to extract from the table
    """
    from .. import SpatialDataFrame
    from arcgis.geometry import _types
    if HASARCPY:
        sql_clause = kwargs.pop('sql_clause', (None,None))
        where_clause = kwargs.pop('where_clause', None)
        sr = kwargs.pop('sr', arcpy.Describe(filename).spatialReference or arcpy.SpatialReference(4326))
        fields = kwargs.pop('fields', None)
        desc = arcpy.Describe(filename)
        if not fields:
            fields = [field.name for field in arcpy.ListFields(filename) \
                      if field.type not in ['Geometry']]

            if hasattr(desc, 'areaFieldName'):
                afn = desc.areaFieldName
                if afn in fields:
                    fields.remove(afn)
            if hasattr(desc, 'lengthFieldName'):
                lfn = desc.lengthFieldName
                if lfn in fields:
                    fields.remove(lfn)
        geom_fields = fields + ['SHAPE@']
        flds = fields + ['SHAPE']
        vals = []
        geoms = []
        geom_idx = flds.index('SHAPE')
        shape_type = desc.shapeType
        default_polygon = _types.Geometry(arcpy.Polygon(arcpy.Array([arcpy.Point(0,0)]* 3)))
        default_polyline = _types.Geometry(arcpy.Polyline(arcpy.Array([arcpy.Point(0,0)]* 2)))
        default_point = _types.Geometry(arcpy.PointGeometry(arcpy.Point()))
        default_multipoint = _types.Geometry(arcpy.Multipoint(arcpy.Array([arcpy.Point()])))
        with arcpy.da.SearchCursor(filename,
                                   field_names=geom_fields,
                                   where_clause=where_clause,
                                   sql_clause=sql_clause,
                                   spatial_reference=sr) as rows:

            for row in rows:
                row = list(row)
                # Prevent curves/arcs
                if row[geom_idx] is None:
                    row.pop(geom_idx)
                    g = {}
                elif row[geom_idx].type in ['polyline', 'polygon']:
                    g = _types.Geometry(row.pop(geom_idx).generalize(0))
                else:
                    g = _types.Geometry(row.pop(geom_idx))
                if g == {}:
                    if shape_type.lower() == 'point':
                        g = default_point
                    elif shape_type.lower() == 'polygon':
                        g = default_polygon
                    elif shape_type.lower() == 'polyline':
                        g = default_point
                    elif shape_type.lower() == 'multipoint':
                        g = default_multipoint
                geoms.append(g)
                vals.append(row)
                del row
            del rows
        df = pd.DataFrame(data=vals, columns=fields)
        sdf = SpatialDataFrame(data=df, geometry=geoms)
        sdf.reset_index(drop=True, inplace=True)
        del df
        if sdf.sr is None:
            if sr is not None:
                sdf.sr = sr
            else:
                sdf.sr = sdf.geometry[sdf.geometry.first_valid_index()].spatialReference
        return sdf
    elif HASARCPY == False and \
         HASPYSHP == True and\
         filename.lower().find('.shp') > -1:
        geoms = []
        records = []
        reader = shapefile.Reader(filename)
        fields = [field[0] for field in reader.fields if field[0] != 'DeletionFlag']
        for r in reader.shapeRecords():
            atr = dict(zip(fields, r.record))
            g = r.shape.__geo_interface__
            g = _geojson_to_esrijson(g)
            geom = _types.Geometry(g)
            atr['SHAPE'] = geom
            records.append(atr)
            del atr
            del r, g
            del geom
        sdf = SpatialDataFrame(records)
        sdf.set_geometry(col='SHAPE')
        sdf.reset_index(inplace=True)
        return sdf
    elif HASARCPY == False and \
         HASPYSHP == False and \
         HASFIONA == True and \
         (filename.lower().find('.shp') > -1 or \
          os.path.dirname(filename).lower().find('.gdb') > -1):
        is_gdb = os.path.dirname(filename).lower().find('.gdb') > -1
        if is_gdb:
            with fiona.drivers():
                from arcgis.geometry import _types
                fp = os.path.dirname(filename)
                fn = os.path.basename(filename)
                geoms = []
                atts = []
                with fiona.open(path=fp, layer=fn) as source:
                    meta = source.meta
                    cols = list(source.schema['properties'].keys())
                    for idx, row in source.items():
                        geoms.append(_types.Geometry(row['geometry']))
                        atts.append(list(row['properties'].values()))
                        del idx, row
                    df = pd.DataFrame(data=atts, columns=cols)
                    return SpatialDataFrame(data=df, geometry=geoms)
        else:
            with fiona.drivers():
                from arcgis.geometry import _types
                geoms = []
                atts = []
                with fiona.open(path=filename) as source:
                    meta = source.meta
                    cols = list(source.schema['properties'].keys())
                    for idx, row in source.items():
                        geoms.append(_types.Geometry(row['geometry']))
                        atts.append(list(row['properties'].values()))
                        del idx, row
                    df = pd.DataFrame(data=atts, columns=cols)
                    return SpatialDataFrame(data=df, geometry=geoms)
    return
#--------------------------------------------------------------------------
def _arcpy_to_featureclass(df, out_name, out_location=None,
                           overwrite=True, out_sr=None,
                           skip_invalid=True):
    """
    """
    import arcgis
    import numpy as np
    import datetime
    from arcpy import da
    from arcgis.features import SpatialDataFrame

    gtype = df.geometry_type.upper()
    gname = df.geometry.name
    df = df.copy()

    if overwrite and \
       arcpy.Exists(os.path.join(out_location, out_name)):
        arcpy.Delete_management(os.path.join(out_location, out_name))
    elif overwrite == False and \
        arcpy.Exists(os.path.join(out_location, out_name)):
        raise Exception("Dataset exists, please provide a new out_name or location.")

    if out_sr is None:
        try:
            if isinstance(df.sr, dict):
                sr = arcgis.geometry.SpatialReference(df.sr).as_arcpy
            elif isinstance(df.sr, arcgis.geometry.SpatialReference):
                sr = df.sr.as_arcpy
        except:
            sr = arcpy.SpatialReference(4326)
    else:
        if isinstance(df.sr, dict):
            sr = arcgis.geometry.SpatialReference(df.sr).as_arcpy
        elif isinstance(df.sr, arcgis.geometry.SpatialReference):
            sr = df.sr.as_arcpy
        elif isinstance(out_sr, arcpy.SpatialReference):
            sr = out_sr
    fc = arcpy.CreateFeatureclass_management(out_path=out_location,
                                             out_name=out_name,
                                             geometry_type=gtype.upper(),
                                             spatial_reference=sr)[0]
    df['JOIN_ID_FIELD_DROP'] = df.index.tolist()
    flds = df.columns.tolist()

    flds.pop(flds.index(gname))
    flds_lower = [f.lower() for f in flds]
    for f in ['objectid', 'oid', 'fid']:
        if f in flds_lower:
            idx = flds_lower.index(f)
            flds.pop(idx)
            flds_lower.pop(idx)
            del idx
        del f
    array = [tuple(row) for row in df[flds].as_matrix()]
    geoms = df.geometry.as_arcpy.tolist()
    dtypes = []

    for idx, a in enumerate(array[0]):
        if isinstance(a,
                      STRING_TYPES):
            dtypes.append((flds[idx], '<U%s' %  df[flds[idx]].map(len).max()))
        elif flds[idx].lower() in ['fid', 'oid', 'objectid']:
            dtypes.append((flds[idx], np.int32))
        elif isinstance(a,
                        (int, np.int32)):
            dtypes.append((flds[idx], np.int64))
        elif isinstance(a,
                        (float, np.float, np.float64)):
            dtypes.append((flds[idx], np.float64))
        elif isinstance(a,
                        (datetime.datetime, pd.datetime)):
            dtypes.append((flds[idx], '<M8[us]'))
        else:
            dtypes.append((flds[idx], type(a)))
        del idx, a

    array = np.array(array, dtype=dtypes)
    del dtypes, flds, flds_lower

    with da.InsertCursor(fc, ['SHAPE@']) as icur:
        for g in geoms:
            if skip_invalid:
                try:
                    icur.insertRow([g])
                except: pass
            else:
                icur.insertRow([g])
    desc = arcpy.Describe(fc)
    oidField = desc.oidFieldName
    del desc
    da.ExtendTable(in_table=fc, table_match_field=oidField,
                   in_array=array, array_match_field='JOIN_ID_FIELD_DROP',
                   append_only=False)
    del df['JOIN_ID_FIELD_DROP']
    return fc
#--------------------------------------------------------------------------
def to_featureclass(df, out_name, out_location=None,
                    overwrite=True, out_sr=None,
                    skip_invalid=True):
    """
    converts a SpatialDataFrame to a feature class

    Parameters:
     :out_location: path to the workspace
     :out_name: name of the output feature class table
     :overwrite: True, the data will be erased then replaced, else the
      table will be appended to an existing table.
     :out_sr: if set, the data will try to reproject itself
     :skip_invalid: if True, the cursor object will not raise an error on
      insertion of invalid data, if False, the first occurence of invalid
      data will raise an exception.
    Returns:
     path to the feature class
    """
    fc = None
    if HASARCPY:
        import arcgis
        cols = []
        dt_idx = []
        invalid_rows = []
        idx = 0
        max_length = None
        if out_location:
            if os.path.isdir(out_location) == False and \
               out_location.lower().endswith('.gdb'):
                out_location = arcpy.CreateFileGDB_management(out_folder_path=os.path.dirname(out_location),
                                                             out_name=os.path.basename(out_location))[0]
            elif os.path.isdir(out_location) == False and \
                 out_name.lower().endswith('.shp'):
                os.makedirs(out_location)
            elif os.path.isfile(out_location) == False and \
                 out_location.lower().endswith('.sde'):
                raise ValueError("The sde connection file does not exist")
        else:
            if out_name.lower().endswith('.shp'):
                out_location = tempfile.gettempdir()
            elif HASARCPY:
                out_location = arcpy.env.scratchGDB
            else:
                out_location = tempfile.gettempdir()
                out_name = out_name + ".shp"
        fc = os.path.join(out_location, out_name)
        df = df.copy() # create a copy so we don't modify the source data.
        if out_name.lower().endswith('.shp'):
            max_length = 10
        for col in df.columns:
            if col.lower() != 'shape':
                if df[col].dtype.type in NUMERIC_TYPES:
                    df[col] = df[col].fillna(0)
                elif df[col].dtype.type in DATETIME_TYPES:
                    dt_idx.append(col)
                else:
                    df.loc[df[col].isnull(), col] = ""
                idx += 1
                col = sanitize_field_name(s=col,
                                          length=max_length)
            cols.append(col)
            del col
        df.columns = cols

        if arcpy.Exists(fc) and \
           overwrite:
            arcpy.Delete_management(fc)
        if arcpy.Exists(fc) ==  False:
            sr = df.sr
            if df.sr is None:
                sr = df['SHAPE'].loc[df['SHAPE'].first_valid_index()].spatial_reference
                if isinstance(sr, dict) and \
                   'wkid' in sr:
                    sr = arcpy.SpatialReference(sr['wkid'])
                elif isinstance(sr, arcpy.SpatialReference):
                    sr = sr
                else:
                    sr = None
            elif df.sr:
                sr = _types.SpatialReference(df.sr).as_arcpy
            elif sr is None:
                sr = df['SHAPE'].loc[df['SHAPE'].first_valid_index()].spatial_reference
                if isinstance(sr, dict) and \
                               'wkid' in sr:
                    sr = arcpy.SpatialReference(sr['wkid'])
                elif isinstance(sr, arcpy.SpatialReference):
                    sr = sr
                else:
                    sr = None
            elif isinstance(sr, dict):
                sr = _types.SpatialReference(sr).as_arcpy
            elif isinstance(sr, _types.SpatialReference):
                sr = df.sr.as_arcpy

            fc = arcpy.CreateFeatureclass_management(out_path=out_location,
                                                     out_name=out_name,
                                                     geometry_type=df.geometry_type.upper(),
                                                     spatial_reference=sr)[0]
        desc = arcpy.Describe(fc)
        oidField = desc.oidFieldName
        col_insert = copy.copy(df.columns).tolist()
        if hasattr(desc, 'areaFieldName'):
            af = desc.areaFieldName.lower()
        else:
            af = None
        if hasattr(desc, 'lengthFieldName'):
            lf = desc.lengthFieldName.lower()
        else:
            lf = None
        col_insert = [f for f in col_insert if f.lower() not in ['oid', 'objectid', 'fid', desc.oidFieldName.lower(), af, lf]]
        df_cols = col_insert.copy()
        lower_col_names = [f.lower() for f in col_insert if f.lower() not in ['oid', 'objectid', 'fid']]
        idx_shp = None

        if oidField.lower() in lower_col_names:
            val = col_insert.pop(lower_col_names.index(oidField.lower()))
            del df[val]
            col_insert = copy.copy(df.columns).tolist()
            lower_col_names = [f.lower() for f in col_insert]
        if hasattr(desc, "areaFieldName") and \
           desc.areaFieldName.lower() in lower_col_names:
            val = col_insert.pop(lower_col_names.index(desc.areaFieldName.lower()))
            del df[val]
            col_insert = copy.copy(df.columns).tolist()
            lower_col_names = [f.lower() for f in col_insert]
        elif 'shape_area' in lower_col_names:
            val = col_insert.pop(lower_col_names.index('shape_area'))
            del df[val]
            col_insert = copy.copy(df.columns).tolist()
            lower_col_names = [f.lower() for f in col_insert]
        if hasattr(desc, "lengthFieldName") and \
           desc.lengthFieldName.lower() in lower_col_names:
            val = col_insert.pop(lower_col_names.index(desc.lengthFieldName.lower()))
            del df[val]
            col_insert = copy.copy(df.columns).tolist()
            lower_col_names = [f.lower() for f in col_insert]
        elif 'shape_length' in lower_col_names:
            val = col_insert.pop(lower_col_names.index('shape_length'))
            del df[val]
            col_insert = copy.copy(df.columns).tolist()
            lower_col_names = [f.lower() for f in col_insert]
        if "SHAPE" in df.columns:
            idx_shp = col_insert.index("SHAPE")
            col_insert[idx_shp] = "SHAPE@"
        existing_fields = [field.name.lower() for field in arcpy.ListFields(fc)]
        for col in col_insert:
            if col.lower() != 'shape@' and \
               col.lower() != 'shape' and \
               col.lower() not in existing_fields:
                try:
                    t = _infer_type(df, col)
                    if t == "TEXT" and out_name.lower().endswith('.shp') == False:
                        l = int(df[col].str.len().max()) or 0
                        if l < 255:
                            l = 255
                        arcpy.AddField_management(in_table=fc, field_name=col,
                                                  field_length=l,
                                                  field_type=_infer_type(df, col))
                    else:
                        arcpy.AddField_management(in_table=fc, field_name=col,
                                              field_type=t)
                except:
                    print('col %s' % col)
        dt_idx = [col_insert.index(col) for col in dt_idx if col in col_insert]
        icur = da.InsertCursor(fc, col_insert)
        for index, row in df[df_cols].iterrows():
            if len(dt_idx) > 0:
                row = row.tolist()
                for i in dt_idx:
                    row[i] = row[i].to_pydatetime()
                    del i
                try:
                    if idx_shp:
                        row[idx_shp] = row[idx_shp].as_arcpy
                    icur.insertRow(row)
                except:
                    invalid_rows.append(index)
                    if skip_invalid == False:
                        raise Exception("Invalid row detected at index: %s" % index)
            else:
                try:
                    row = row.tolist()
                    if isinstance(idx_shp, int):
                        row[idx_shp] = row[idx_shp].as_arcpy
                    icur.insertRow(row)
                except:
                    invalid_rows.append(index)
                    if skip_invalid == False:
                        raise Exception("Invalid row detected at index: %s" % index)

            del row
        del icur
        if len(invalid_rows) > 0:
            t = ",".join([str(r) for r in invalid_rows])
            _log.warning('The following rows could not be written to the table: %s' % t)
    elif HASARCPY == False and \
         HASPYSHP:
        return _pyshp_to_shapefile(df=df,
                                   out_path=out_location,
                                   out_name=out_name)
    else:
        raise Exception("Cannot Export the data without ArcPy or PyShp modules. "+ \
                        "Please install them and try again.")
    return fc
#--------------------------------------------------------------------------
def _infer_type(df, col):
    """
    internal function used to get the datatypes for the feature class if
    the dataframe's _field_reference is NULL or there is a column that does
    not have a dtype assigned to it.

    Input:
     dataframe - spatialdataframe object
    Ouput:
      field type name
    """
    nn = df[col].notnull()
    nn = list(df[nn].index)
    if len(nn) > 0:
        val = df[col][nn[0]]
        if isinstance(val, six.string_types):
            return "TEXT"
        elif isinstance(val, tuple(list(six.integer_types) + [np.int32])):
            return "INTEGER"
        elif isinstance(val, (float, np.int64 )):
            return "FLOAT"
        elif isinstance(val, datetime):
            return "DATE"
    return "TEXT"
#--------------------------------------------------------------------------
def _geojson_to_esrijson(geojson):
    """converts the geojson spec to esri json spec"""
    if geojson['type'] in ['Polygon', 'MultiPolygon']:
        return {
            'rings' : geojson['coordinates']
        }
    elif geojson['type'] == "Point":
        return {
            "x" : geojson['coordinates'][0],
            "y" : geojson['coordinates'][1]
        }
    elif geojson['type'] == "MultiPoint":
        return {
            "points" : geojson['coordinates'],
        }
    elif geojson['type'] in ['LineString', 'MultiLineString']:
        return {
            "paths" : geojson['coordinates'],
        }
    return geojson
#--------------------------------------------------------------------------
def _geometry_to_geojson(geom):
    """converts the esri json spec to geojson"""
    if 'rings' in geom and \
       len(geom['rings']) == 1:
        return {
            'type' : "Polygon",
            "coordinates" : geom['rings']
        }
    elif 'rings' in geom and \
       len(geom['rings']) > 1:
        return {
            'type' : "MultiPolygon",
            "coordinates" : geom['rings']
        }
    elif geom['type'] == "Point":
        return {
            "coordinates" : [geom['x'], geom['y']],
            "type" : "Point"
        }
    elif geom['type'] == "MultiPoint":
        return {
            "coordinates" : geom['points'],
            'type' : "MultiPoint"
        }
    elif geom['type'].lower() == "polyline" and \
         len(geom['paths']) <= 1:
        return {
            "coordinates" : geom['paths'],
            'type' : "LineString"
        }
    elif geom['type'].lower() == "polyline" and \
         len(geom['paths']) > 1:
        return {
            "coordinates" : geom['paths'],
            'type' : "MultiLineString"
        }
    return geom


"""
The arcgis.geoenrichment module  enables you to answer questions about locations that you can't answer with maps alone.

GeoEnrichment provides the ability to get facts about a location or area. Using GeoEnrichment, you can get information
about the people, places, and businesses in a specific area or within a certain distance or drive time from a location.
It enables you to query and use information from a large collection of data sets including population, income, housing,
consumer behavior, and the natural environment.

For example: What kind of people live here? What do people like to do in this area? What are their habits and
lifestyles? What kind of businesses are in this area?

The enrich() method to can be used retrieve demographics and other relevant characteristics associated with the area
surrounding the requested places. You can also use the arcgis.geoenrichment module to obtain additional geographic
context (for example, the ZIP Code of a location) and geographic boundaries (for example, the geometry for a drive-time
service area).

Site analysis is a popular application of this type of data enrichment. For example, GeoEnrichment can be leveraged to
study the population that would be affected by the development of a new community center within their neighborhood.
With the enrich() method, the proposed site can be submitted, and the demographics and other relevant characteristics
associated with the area around the site will be returned.
"""
from .enrichment import *



""" The portalpy module for working with the ArcGIS Online and Portal APIs."""
from __future__ import absolute_import
import io
import os
import re
import sys
import json
import uuid
import zlib
import shutil
import logging
import datetime
import tempfile
import mimetypes
import unicodedata
try:
    #PY2
    from cStringIO import StringIO
except ImportError:
    #PY3
    from io import StringIO
from io import BytesIO
from collections import OrderedDict

import six
from six.moves.urllib_parse import urlparse, urlunparse, parse_qsl
from six.moves.urllib_parse import quote, unquote, urlunsplit
from six.moves.urllib_parse import urlencode, urlsplit
from six.moves.urllib.error import HTTPError
from six.moves.urllib import request
from six.moves import http_cookiejar as cookiejar
from six.moves import http_client
from arcgis._impl.connection import _ArcGISConnection
from arcgis.gis import GIS
class Error(Exception): pass

__version__ = '2.0'
_log = logging.getLogger(__name__)
########################################################################
class MultiPartForm(object):
    """Accumulate the data to be used when posting a form."""
    PY2 = sys.version_info[0] == 2
    PY3 = sys.version_info[0] == 3
    files = []
    form_fields = []
    boundary = None
    form_data = ""
    #----------------------------------------------------------------------
    def __init__(self, param_dict=None, files=None):
        if param_dict is None:
            param_dict = {}
        if files is None:
            files = {}
        self.boundary = None
        self.files = []
        self.form_data = ""
        if len(self.form_fields) > 0:
            self.form_fields = []

        if len(param_dict) == 0:
            self.form_fields = []
        else:
            for k,v in param_dict.items():
                self.form_fields.append((k,v))
                del k,v
        if isinstance(files, list):
            if len(files) == 0:
                self.files = []
            else:
                for key, filePath, fileName in files:
                    self.add_file(fieldname=key,
                                  filename=fileName,
                                  filePath=filePath,
                                  mimetype=None)
        elif isinstance(files, dict):
            for key, filepath in files.items():
                self.add_file(fieldname=key,
                              filename=os.path.basename(filepath),
                              filePath=filepath,
                              mimetype=None)
                del key, filepath
        self.boundary = "-%s" % self._make_boundary()
    #----------------------------------------------------------------------
    def get_content_type(self):
        return 'multipart/form-data; boundary=%s' % self.boundary
    #----------------------------------------------------------------------
    def add_field(self, name, value):
        """Add a simple field to the form data."""
        self.form_fields.append((name, value))
    #----------------------------------------------------------------------
    def _make_boundary(self):
        """ creates a boundary for multipart post (form post)"""
        if six.PY2:
            return '-===============%s==' % uuid.uuid4().get_hex()
        elif six.PY3:
            return '-===============%s==' % uuid.uuid4().hex
        else:
            from random import choice
            digits = "0123456789"
            letters = "abcdefghijklmnopqrstuvwxyz"
            return '-===============%s==' % ''.join(choice(letters + digits) \
                                                    for i in range(15))
    #----------------------------------------------------------------------
    def add_file(self, fieldname, filename, filePath, mimetype=None):
        """Add a file to be uploaded.
        Inputs:
           fieldname - name of the POST value
           fieldname - name of the file to pass to the server
           filePath - path to the local file on disk
           mimetype - MIME stands for Multipurpose Internet Mail Extensions.
             It's a way of identifying files on the Internet according to
             their nature and format. Default is None.
        """
        body = filePath
        if mimetype is None:
            mimetype = mimetypes.guess_type(filename)[0] or 'application/octet-stream'
        self.files.append((fieldname, filename, mimetype, body))
    #----------------------------------------------------------------------
    @property
    def make_result(self):
        if self.PY2:
            self._2()
        elif self.PY3:
            self._3()
        return self.form_data
    #----------------------------------------------------------------------
    def _2(self):
        """python 2.x version of formatting body data"""
        boundary = self.boundary
        buf = StringIO()
        for (key, value) in self.form_fields:
            buf.write('--%s\r\n' % boundary)
            buf.write('Content-Disposition: form-data; name="%s"' % key)
            buf.write('\r\n\r\n%s\r\n' % value)
        for (key, filename, mimetype, filepath) in self.files:
            if os.path.isfile(filepath):
                buf.write('--{boundary}\r\n'
                          'Content-Disposition: form-data; name="{key}"; '
                          'filename="{filename}"\r\n'
                          'Content-Type: {content_type}\r\n\r\n'.format(
                              boundary=boundary,
                              key=key,
                              filename=filename,
                              content_type=mimetype))
                with open(filepath, "rb") as f:
                    shutil.copyfileobj(f, buf)
                buf.write('\r\n')
        buf.write('--' + boundary + '--\r\n\r\n')
        buf = buf.getvalue()
        self.form_data = buf
    #----------------------------------------------------------------------
    def _3(self):
        """ python 3 method"""
        boundary = self.boundary
        buf = BytesIO()
        textwriter = io.TextIOWrapper(
            buf, 'utf8', newline='', write_through=True)

        for (key, value) in self.form_fields:
            textwriter.write(
                '--{boundary}\r\n'
                'Content-Disposition: form-data; name="{key}"\r\n\r\n'
                '{value}\r\n'.format(
                    boundary=boundary, key=key, value=value))
        for(key, filename, mimetype, filepath) in self.files:
            if os.path.isfile(filepath):
                textwriter.write(
                    '--{boundary}\r\n'
                    'Content-Disposition: form-data; name="{key}"; '
                    'filename="{filename}"\r\n'
                    'Content-Type: {content_type}\r\n\r\n'.format(
                        boundary=boundary, key=key, filename=filename,
                        content_type=mimetype))
                with open(filepath, "rb") as f:
                    shutil.copyfileobj(f, buf)
                textwriter.write('\r\n')
        textwriter.write('--{}--\r\n\r\n'.format(boundary))
        self.form_data = buf.getvalue()
########################################################################
class HTTPSClientAuthHandler(request.HTTPSHandler):
    def __init__(self, key, cert):
        request.HTTPSHandler.__init__(self)
        self.key = key
        self.cert = cert
    def https_open(self, req):
        #Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)
    def getConnection(self, host, timeout=300):
        return  http_client.HTTPSConnection(host,
                                            key_file=self.key,
                                            cert_file=self.cert,
                                            timeout=timeout)

def jsonize_dict(val):
    if isinstance(val, (dict, list)):
        return json.dumps(val)
    else:
        return val

class ServerConnection(object):
    """

    =====================     ====================================================================
    **Argument**              **Description**
    ---------------------     --------------------------------------------------------------------
    baseurl                   optional string, the root URL to a site.
                              Example: https://mysite.com/arcgis
    ---------------------     --------------------------------------------------------------------
    tokenurl                  optional string. Used when a site if federated or when the token
                              URL differs from the site's baseurl.  If a site is federated, the
                              token URL will return as the Portal token and ArcGIS Server users
                              will not validate correctly.
    ---------------------     --------------------------------------------------------------------
    username                  optional string, login username for BUILT-IN security
    ---------------------     --------------------------------------------------------------------
    password                  optional string, a secret word or phrase that must be used to gain
                              access to the account above.
    ---------------------     --------------------------------------------------------------------
    key_file                  optional string, path to PKI ket file
    ---------------------     --------------------------------------------------------------------
    cert_file                 optional string, path to PKI cert file
    ---------------------     --------------------------------------------------------------------
    proxy_host                optional string, web address to the proxy host

                              Example: proxy.mysite.com
    ---------------------     --------------------------------------------------------------------
    proxy_port                optional integer, default is 80. The port where the proxy resided on
    ---------------------     --------------------------------------------------------------------
    expiration                optional integer. The Default is 60. This is the length of time a
                              token is valid for.
                              Example 1440 is one week.
    ---------------------     --------------------------------------------------------------------
    all_ssl                   optional boolean. The default is False. If True, all calls will be
                              made over HTTPS instead of HTTP.
    ---------------------     --------------------------------------------------------------------
    portal_connection         optional GIS. This is used when a site is federated. It is the
                              ArcGIS Online or Portal GIS object used.
    =====================     ====================================================================


    """
    baseurl = None
    key_file = None
    cert_file = None
    all_ssl = None
    proxy_host = None
    proxy_port = None
    _token = None
    _product = None
    _referer = None
    _useragent = None
    _parsed_org_url = None
    _username = None
    _password = None
    _auth = None
    _tokenurl = None
    _token = None
    _server_token = None
    _connection = None
    _portal_connection = None
    _TOKEN_TIME = None
    _REFRESH_WHEN = None
    _handlers = None
    def __init__(self, baseurl=None, tokenurl=None, username=None,
                 password=None, key_file=None, cert_file=None,
                 expiration=60, all_ssl=False, referer=None,
                 proxy_host=None, proxy_port=None,
                 portal_connection=None, **kwargs):
        """ The ServerConnection constructor. Requires URL and optionally username/password. """
        self._verify_cert = kwargs.pop('verify_cert', True)
        self.baseurl = baseurl
        if self._tokenurl is None and \
           portal_connection is None:
            parsed = urlparse(baseurl)
            tokenurl = "{scheme}://{netloc}/{wa}/admin/generateToken".format(
                scheme=parsed.scheme,
                netloc=parsed.netloc,
                wa=parsed.path[1:].split('/')[0]
            )
        self._tokenurl = tokenurl
        self._username = username
        self._password = password
        self._expiration = expiration
        self.all_ssl = all_ssl
        if referer:
            self._referer = referer
        else:
            self._referer = urlparse(baseurl).netloc
        self.key_file = key_file
        self.cert_file = cert_file
        self.proxy_host = proxy_host
        self.proxy_port = proxy_port
        if portal_connection:
            if isinstance(portal_connection, _ArcGISConnection):
                self._portal_connection = portal_connection # second connection
            elif hasattr(self._portal_connection, '_con') and \
                 getattr(self._portal_connection, "_con") is not None:
                self._portal_connection = self._portal_connection._con
            elif isinstance(portal_connection, GIS):
                self._portal_connection = portal_connection._con
            else:
                raise ValueError("A portal_connection object be of type _ArcGISConnection or GIS")
        if self._portal_connection:
            self._product = "FEDERATED_SERVER"
        else:
            self._product = "SERVER"

        # Setup the referer and user agent
        if not referer:
            referer = urlparse(baseurl).netloc
        self._useragent = 'geosaurus/' + __version__
        parsed_url = urlparse(self.baseurl)
        self._parsed_org_url = urlunparse((parsed_url[0], parsed_url[1], "", "", "", ""))

        if cert_file is not None and key_file is not None:
            self._auth = "PKI"
        elif username is not None and password is not None:
            self._auth = "BUILTIN" # or "BASICAUTH" (LDAP) or NTLM or Kerberos (login sets this up)
        elif portal_connection:
            self._auth = "PORTAL_TOKEN"
            self.token
        else:
            self._auth = "ANON"

        # Login if credentials were provided
        if username and password:
            self.login(username, password, expiration)
        elif username or password:
            _log.warning('Both username and password required for login')
    #----------------------------------------------------------------------
    @property
    def product(self):
        if self._product is None:
            self._product = self._check_product()
        return self._product
    #----------------------------------------------------------------------
    @property
    def portal_connection(self):
        """gets/sets an additional connection object to get a token from"""
        return self._portal_connection
    #----------------------------------------------------------------------
    @portal_connection.setter
    def portal_connection(self, value):
        """gets/sets an additional connection object to get a token from"""
        if self._portal_connection != value:
            self._portal_connection = value
            self._token = None
            self._server_token = None
    #----------------------------------------------------------------------
    @property
    def token(self):
        """gets/sets the token"""
        if self._token:
            if self._REFRESH_WHEN and \
               datetime.datetime.now() > self._REFRESH_WHEN:
                self._token = None
                self.token
            return self._token
        if self._portal_connection and \
           self._server_token is None:
            if self._portal_connection.product == "AGO":
                return self._portal_connection.token
            parsed = urlparse(self.baseurl)
            adminURL = "https://%s/%s/admin" % (parsed.netloc, urlparse(self.baseurl).path[1:].split('/')[0])
            self._token = self.portal_connection.generate_portal_server_token(serverUrl=adminURL)
            self._REFRESH_WHEN = datetime.datetime.now() + datetime.timedelta(seconds=self._expiration)
            return self._token
        elif self._portal_connection and self._server_token:
            return self._server_token
        elif self._username and self._password:
            self.login(username=self._username,
                       password=self._password,
                       expiration=60)
            self._REFRESH_WHEN = datetime.datetime.now() + datetime.timedelta(seconds=self._expiration)
            return self._token
        return None
    #----------------------------------------------------------------------
    @token.setter
    def token(self, value):
        """gets/sets the token"""
        if self._token != value:
            self._token = value
    #----------------------------------------------------------------------
    def generate_token(self, username, password, expiration=60):
        """ Generates and returns a new token, but doesn't re-login. """
        if self.product == "SERVER":
            postdata = { 'username': username,
                         'password': password,
                         'client': 'requestip',
                         'expiration': expiration,
                         'f': 'json' }
        elif self.product == "FEDERATED_SERVER" and \
             self._portal_connection:
            parsed = urlparse(self.baseurl)
            adminURL = "%s://%s/%s/admin" % (parsed.scheme, parsed.netloc, urlparse(self.baseurl).path[1:].split('/')[0])
            token =  self.portal_connection.generate_portal_server_token(serverUrl=adminURL)
            return token
        else: # Assume username/password BUITIN
            postdata = { 'username': username, 'password': password,
                         'client': 'referer', 'referer': self._referer,
                         'expiration': expiration, 'f': 'json' }
        if self._tokenurl is None:
            if self.baseurl.endswith('/'):
                resp = self.post('generateToken', postdata,
                                 ssl=urlparse(self._tokenurl).scheme == 'https', add_token=False)
            else:
                resp = self.post('/generateToken', postdata,
                                 ssl=urlparse(self._tokenurl).scheme == 'https', add_token=False)
        else:

            resp = self.post(path=self._tokenurl, postdata=postdata,
                             ssl=urlparse(self._tokenurl).scheme == "https", add_token=False)
        if resp:
            return resp.get('token')
    #----------------------------------------------------------------------
    def login(self, username, password, expiration=60):
        """ Logs into the portal using username/password. """
        try:
            newtoken = self.generate_token(username,
                                           password, expiration)
            self._REFRESH_WHEN = datetime.datetime.now() + datetime.timedelta(seconds=expiration)
            if newtoken:
                self._token = newtoken
                self._username = username
                self._password = password
                self._expiration = expiration
                self._auth = "BUILTIN"
            return newtoken
        except HTTPError as err:
            if err.code == 401: # using basic authentication
                self._auth = "BASICAUTH"
            else:
                raise
    #----------------------------------------------------------------------
    def relogin(self, expiration=None):
        """ Re-authenticates with the portal using the same username/password. """
        if not expiration:
            expiration = self._expiration
        return self.login(self._username, self._password, expiration)
    #----------------------------------------------------------------------
    def logout(self):
        """ Logs out of the portal. """
        self._token = None
        self._server_token = None
    #----------------------------------------------------------------------
    @property
    def is_logged_in(self):
        """ Returns true if logged into the portal. """
        return self._token is not None or self._server_token is not None
    #----------------------------------------------------------------------
    def _mainType(self, resp):
        """ gets the main type from the response object"""
        if six.PY2:
            return resp.headers.maintype
        elif six.PY3:
            return resp.headers.get_content_maintype()
        else:
            return None
    #----------------------------------------------------------------------
    def _check_product(self):
        """
        determines if the product is portal, arcgis online or arcgis server
        """
        baseurl = self.baseurl
        if self.portal_connection:
            return "FEDERATED_SERVER"
        return "SERVER"
    #----------------------------------------------------------------------
    def _get_file_name(self, contentDisposition,
                       url, ext=".unknown"):
        """ gets the file name from the header or url if possible """
        if six.PY2:
            if contentDisposition is not None:
                return re.findall(r'filename[^;=\n]*=(([\'"]).*?\2|[^;\n]*)',
                                  contentDisposition.strip().replace('"', ''))[0][0]
            elif os.path.basename(url).find('.') > -1:
                return os.path.basename(url)
        elif six.PY3:
            if contentDisposition is not None:
                p = re.compile(r'filename[^;=\n]*=(([\'"]).*?\2|[^;\n]*)')
                return p.findall(contentDisposition.strip().replace('"', ''))[0][0]
            elif os.path.basename(url).find('.') > -1:
                return os.path.basename(url)
        return "%s.%s" % (uuid.uuid4().get_hex(), ext)
    #----------------------------------------------------------------------
    def _process_response(self, resp, out_folder=None,  file_name=None):
        """ processes the response object"""
        CHUNK = 4056
        maintype = self._mainType(resp)
        contentDisposition = resp.headers.get('content-disposition')
        contentType = resp.headers.get('content-type')
        contentLength = resp.headers.get('content-length')
        if contentType.find('application/json;') == -1 and \
           (maintype.lower() in ('image',
                                'application',
                                'application/x-zip-compressed') or \
           contentType == 'application/x-zip-compressed' or \
           (contentDisposition is not None and \
            contentDisposition.lower().find('attachment;') > -1)):
            fname = self._get_file_name(
                contentDisposition=contentDisposition,
                url=resp.geturl()).split('?')[0]
            if out_folder is None:
                out_folder = tempfile.gettempdir()
            if contentLength is not None:
                max_length = int(contentLength)
                if max_length < CHUNK:
                    CHUNK = max_length
            if file_name is None:
                file_name = os.path.join(out_folder, fname)
            else:
                file_name = os.path.join(out_folder, file_name)
            with open(file_name, 'wb') as writer:
                for data in self._chunk(response=resp):
                    writer.write(data)
                    del data
                del writer
            return file_name
        else:
            read = ""
            if file_name and out_folder:
                f_n_path = os.path.join(out_folder, file_name)
                with open(f_n_path, 'wb') as writer:
                    for data in self._chunk(response=resp, size=4096):
                        if six.PY3 == True:
                            writer.write(data)
                        else:
                            writer.write(data)
                        del data
                    writer.flush()
                return f_n_path
            else:
                for data in self._chunk(response=resp, size=4096):
                    if six.PY3 == True:
                        if read == "":
                            read = data
                        else:
                            read += data
                    else:
                        read += data
                    del data
            if six.PY3 and \
               len(read) > 0:
                try:
                    read = read.decode("utf-8").strip()
                except:
                    pass
            try:
                return read.strip()
            except:
                return read
        return ""
    #----------------------------------------------------------------------
    def _chunk(self, response, size=4096):
        """
        downloads a web response in pieces to ensure there are no
        memory issues.
        """
        method = response.headers.get("content-encoding")
        if method == "gzip":
            d = zlib.decompressobj(16+zlib.MAX_WBITS)
            b = response.read(size)
            while b:
                data = d.decompress(b)
                yield data
                b = response.read(size)
                del data
        else:
            while True:
                chunk = response.read(size)
                if not chunk: break
                yield chunk
    # ----------------------------------------------------------------------
    def get(self, path, params=None, ssl=False,
            compress=True, try_json=True, is_retry=False,
            use_ordered_dict=False, out_folder=None,
            file_name=None, force_bytes=False, **kwargs):
        """ Returns result of an HTTP GET. Handles token timeout and all SSL mode."""
        path = quote(path, ':/%')
        url = path
        if url.lower().find("https://") > -1 or\
           url.lower().find("http://") > -1:
            url = path
        elif len(url) == 0:
            url = self.baseurl
        elif (len(url) > 0 and url[0] == '/' ) == False and \
             self.baseurl.endswith('/') == False:
            url = "/{path}".format(path=url)

        if not url.startswith('http://') and \
           not url.startswith('https://'):
            url = self.baseurl + url
        if ssl or self.all_ssl:
            url = url.replace('http://', 'https://')

        # Add the token if logged in
        if params is None:
            params = {}
        if try_json:
            params['f'] = 'json'
        if self.is_logged_in:
            params['token'] = self.token
        if len(params.keys()) > 0:
            params = {k: jsonize_dict(v) for k, v in params.items()}
            url = "{url}?{params}".format(url=url,
                                          params=urlencode(params))
            #url = self._url_add_token(url, self.token)

        _log.debug('REQUEST (get): ' + url)

        try:
            # Send the request and read the response
            headers = [('User-Agent', self._useragent)]
            if self._referer:# and \
                #self._auth.lower() != 'pki':
                headers.append(('Referer', self._referer))

            if compress:
                headers.append(('Accept-encoding', 'gzip'))
            if self._handlers is None:
                self._handlers = self.get_handlers()
            opener = request.build_opener(*self._handlers)
            opener.addheaders = headers

            req = request.Request(url,
                                  headers={i[0] : i[1] for i in headers})
            resp = opener.open(req)
            resp_data = self._process_response(resp,
                                               out_folder=out_folder,
                                               file_name=file_name)
            #  if the response is a file saved to disk, return it.

            if (len(resp_data) < 32767) and os.path.isfile(resp_data):
                if force_bytes:
                    return open(resp_data, 'rb').read()
                return resp_data
            # If we're not trying to parse to JSON, return response as is
            if not try_json:
                return resp_data

            try:
                if use_ordered_dict:
                    resp_json = json.loads(resp_data,
                                           object_pairs_hook=OrderedDict)
                else:
                    resp_json = json.loads(resp_data)

                # Check for errors, and handle the case where the token timed
                # out during use (and simply needs to be re-generated)
                try:
                    if resp_json.get('error', None):
                        errorcode = resp_json['error']['code']
                        if errorcode == 498 and not is_retry:
                            _log.info('Token expired during get request, ' \
                                      + 'fetching a new token and retrying')
                            newtoken = self.relogin()
                            newpath = self._url_add_token(path, newtoken)
                            return self.get(path=newpath, params=params, ssl=ssl, compress=compress,
                                            try_json=try_json, is_retry=True)
                        elif errorcode == 498:
                            raise RuntimeError('Invalid token')
                        elif errorcode == 403:
                            message = resp_json['error']['message'] if 'message' in resp_json['error'] else ''
                            if message == "SSL Required":
                                return self.get(path=path, params=params, ssl=True, compress=compress,
                                                try_json=try_json, is_retry=True)

                        self._handle_json_error(resp_json['error'])
                        return None
                except AttributeError:
                    # Top-level JSON object isnt a dict, so can't have an error
                    pass

                # If the JSON parsed correctly and there are no errors,
                # return the JSON
                #if 'status' in resp_json: # FOR DEMO TODO REMOVE ME
                #    return resp_json['status'] == 'success'
                return resp_json

            # If we couldnt parse the response to JSON, return it as is
            except ValueError:
                return resp

        # If we got an HTTPError when making the request check to see if it's
        # related to token timeout, in which case, regenerate a token
        except HTTPError as e:
            if e.code == 498 and not is_retry:
                _log.info('Token expired during get request, fetching a new ' \
                          + 'token and retrying')
                self.logout()
                newtoken = self.relogin()
                newpath = self._url_add_token(path, newtoken)
                return self.get(newpath, ssl, try_json, is_retry=True)
            elif e.code == 498:
                raise RuntimeError('Invalid token')
            else:
                raise e
    #----------------------------------------------------------------------
    def _ensure_dir(self, f):
        if not os.path.exists(f):
            os.makedirs(f)
    #----------------------------------------------------------------------
    def _url_add_token(self, url, token):

        # Parse the URL and query string
        urlparts = urlparse(url)
        qs_list = parse_qsl(urlparts.query)

        # Update the token query string parameter
        replaced_token = False
        new_qs_list = []
        for qs_param in qs_list:
            if qs_param[0] == 'token':
                qs_param = ('token', token)
                replaced_token = True
            new_qs_list.append(qs_param)
        if not replaced_token:
            new_qs_list.append(('token', token))

        # Rebuild the URL from parts and return it
        return urlunparse((urlparts.scheme, urlparts.netloc,
                           urlparts.path, urlparts.params,
                           urlencode(new_qs_list),
                           urlparts.fragment))
    #----------------------------------------------------------------------
    def get_handlers(self):
        from urllib.request import HTTPRedirectHandler
        redirect_handler = HTTPRedirectHandler()
        redirect_handler.max_redirections = 30
        redirect_handler.max_repeats = 30
        handlers = [redirect_handler]
        if self._username and self._password:

            passman = request.HTTPPasswordMgrWithDefaultRealm()
            passman.add_password(None,
                                 self._parsed_org_url,
                                 self._username,
                                 self._password)
            handlers.append(request.HTTPBasicAuthHandler(passman))
            passman = request.HTTPPasswordMgrWithDefaultRealm()
            passman.add_password(None,
                                 self._parsed_org_url,
                                 self._username,
                                 self._password)
            handlers.append(request.HTTPDigestAuthHandler(passman))
            if os.name == 'nt':
                try:
                    from arcgis._impl.common._iwa import NtlmSspiAuthHandler, KerberosSspiAuthHandler

                    auth_krb = KerberosSspiAuthHandler()
                    handlers.append(auth_krb)

                    try:
                        auth_NTLM = NtlmSspiAuthHandler()
                        handlers.append(auth_NTLM)
                    except:
                        pass



                except Error as err:
                    _log.error("winkerberos packages is required for IWA authentication (NTLM and Kerberos).")
                    _log.error("Please install it:\n\tconda install winkerberos")
                    _log.error(str(err))
            else:
                _log.error('The GIS uses Integrated Windows Authentication which is currently only supported on the Windows platform')


        if self._auth == "PKI" or \
           (self.cert_file is not None and self.key_file is not None):
            handlers.append(HTTPSClientAuthHandler(self.key_file, self.cert_file))
        elif self._portal_connection and \
             self._portal_connection.cert_file is not None and \
             self._portal_connection.key_file is not None:
            handlers.append(HTTPSClientAuthHandler(self._portal_connection.key_file,
                                                   self._portal_connection.cert_file))

        cj = cookiejar.CookieJar()

        if self.proxy_host: # Simple Proxy Support
            from urllib.request import ProxyHandler
            if self.proxy_port is None:
                self.proxy_port = 80
            proxies = {"http":"http://%s:%s" % (self.proxy_host, self.proxy_port),
                       "https":"https://%s:%s" % (self.proxy_host, self.proxy_port)}
            proxy_support = ProxyHandler(proxies)
            handlers.append(proxy_support)

        handlers.append(request.HTTPCookieProcessor(cj))
        return handlers
    #----------------------------------------------------------------------
    def post(self, path, postdata=None, files=None, ssl=False, compress=True,
             is_retry=False, use_ordered_dict=False, add_token=True, verify_cert=True,
             token=None, **kwargs):
        """ Returns result of an HTTP POST. Supports Multipart requests."""
        #if path.find(" ") > -1:
        path = quote(path, ':/%')
        out_folder = kwargs.pop('out_folder', tempfile.gettempdir())

        url = path
        if url.lower().find("https://") > -1 or\
           url.lower().find("http://") > -1:
            url = path
        elif len(url) == 0:
            url = self.baseurl
        elif (len(url) > 0 and url[0] == '/' ) == False and \
           self.baseurl.endswith('/') == False:
            url = "/{path}".format(path=url)

        if not url.startswith('http://') and \
           not url.startswith('https://'):
            url = self.baseurl + url

        if ssl or self.all_ssl:
            url = url.replace('http://', 'https://')
        if verify_cert == False:
            import ssl
            ssl._create_default_https_context = ssl._create_unverified_context
        # Add the token if logged in
        if add_token:
            if self.is_logged_in:
                postdata['token'] = self.token
        if token:
            postdata['token'] = token
        if _log.isEnabledFor(logging.DEBUG):
            msg = 'REQUEST: ' + url + ', ' + str(postdata)
            if files:
                msg += ', files=' + str(files)
            _log.debug(msg)

        # If there are files present, send a multipart request
        if files:
            #parsed_url = urlparse(url)
            mpf = MultiPartForm(param_dict=postdata, files=files)
            req = request.Request(url)
            body = mpf.make_result
            req.add_header('User-agent', self._useragent)
            req.add_header('Content-type', mpf.get_content_type())
            req.add_header('Content-length', len(body))
            req.data = body
            headers = [('User-Agent', self._useragent),
                       ('Content-type', mpf.get_content_type()),
                       ('Content-length', len(body))]
            if self._referer and \
               self._auth.lower() != 'pki':
                headers.append(('Referer', self._referer))
            if compress:
                headers.append(('Accept-encoding', 'gzip'))
            if self._handlers is None:
                self._handlers = self.get_handlers()
            #handlers = self.get_handlers()
            opener = request.build_opener(*self._handlers)

            opener.addheaders = headers

            resp = opener.open(req)
            resp_data = self._process_response(resp)
        # Otherwise send a normal HTTP POST request
        else:
            encoded_postdata = None
            if postdata:
                postdata = {k: jsonize_dict(v) for k, v in postdata.items()}
                encoded_postdata = urlencode(postdata)
            headers = [('User-Agent', self._useragent)]
            if self._referer and \
                self._auth.lower() != 'pki':
                headers.append(('Referer', self._referer))
            if compress:
                headers.append(('Accept-encoding', 'gzip'))
            if self._handlers is None:
                self._handlers = self.get_handlers()

            opener = request.build_opener(*self._handlers)
            opener.addheaders = headers
            #request.install_opener(opener)
            req = request.Request(url,
                                  data=encoded_postdata.encode('utf-8'),
                                  headers={i[0] : i[1] for i in headers})
            resp = opener.open(req)#request.urlopen(req)
            resp_data = self._process_response(resp,
                                               out_folder=out_folder,
                                               file_name=None)
        # Parse the response into JSON
        if _log.isEnabledFor(logging.DEBUG):
            _log.debug('RESPONSE: ' + url + ', ' + resp_data)
        #print(resp_data);
        if use_ordered_dict:
            resp_json = json.loads(resp_data, object_pairs_hook=OrderedDict)
        else:
            resp_json = json.loads(resp_data)


        # Check for errors, and handle the case where the token timed out
        # during use (and simply needs to be re-generated)
        try:
            if 'error' in resp_json or \
               ('status' in resp_json and \
                resp_json.get('status', None) != "success"):

                errorcode = resp_json['code'] if 'code' in resp_json else 0
                if errorcode == 498 and not is_retry:
                    _log.info('Token expired during post request, fetching a new '
                              + 'token and retrying')
                    self.logout()
                    newtoken = self.relogin()
                    postdata['token'] = newtoken
                    return self.post(path, postdata, files, ssl, compress,
                                     is_retry=True)
                elif errorcode == 498:
                    raise RuntimeError('Invalid token')
                elif errorcode == 403:
                    message = resp_json['error']['message'] if 'message' in resp_json['error'] else ''
                    if message == "SSL Required":
                        return self.post(path, postdata, files, ssl=True, compress=compress, token=token,
                                         verify_cert=verify_cert, is_retry=True)

                if 'status' in resp_json:
                    self._handle_json_error(resp_json, errorcode)
                else:
                    self._handle_json_error(resp_json['error'], errorcode)
                return None
        except AttributeError:
            # Top-level JSON object isnt a dict, so can't have an error
            pass
        return resp_json
    #----------------------------------------------------------------------
    def _get_content_type(self, filename):
        return mimetypes.guess_type(filename)[0] or 'application/octet-stream'
    #----------------------------------------------------------------------
    def _handle_json_error(self, error, errorcode):
        if 'messages' in error:
            errormessage = ""
        else:
            errormessage = error.get('message', 'Unknown Error')
        _log.error(errormessage)
        if 'details' in error and error['details'] is not None:
            for errordetail in error['details']:
                errormessage = errormessage + "\n" + errordetail
                _log.error(errordetail)
        elif 'messages' in error and error['messages'] is not None:
            for errordetail in error['messages']:
                errormessage = errormessage + "\n" + errordetail
        errormessage = errormessage + "\n(Error Code: " + str(errorcode) +")"
        raise RuntimeError(errormessage)
########################################################################
class _StrictURLopener(request.FancyURLopener):
    def http_error_default(self, url, fp, errcode, errmsg, headers):
        if errcode != 200:
            raise HTTPError(url, errcode, errmsg, headers, fp)

def _normalize_url(url, charset='utf-8'):
    """ Normalizes a URL. Based on http://code.google.com/p/url-normalize."""
    def _clean(string):
        string = str(unquote(string), 'utf-8', 'replace')
        return unicodedata.normalize('NFC', string).encode('utf-8')

    default_port = {
        'ftp': 21,
        'telnet': 23,
        'http': 80,
        'gopher': 70,
        'news': 119,
        'nntp': 119,
        'prospero': 191,
        'https': 443,
        'snews': 563,
        'snntp': 563,
    }

    # if there is no scheme use http as default scheme
    if url[0] not in ['/', '-'] and ':' not in url[:7]:
        url = 'http://' + url


    # shebang urls support
    url = url.replace('#!', '?_escaped_fragment_=')

    # splitting url to useful parts
    scheme, auth, path, query, fragment = urlsplit(url.strip())
    (userinfo, host, port) = re.search('([^@]*@)?([^:]*):?(.*)', auth).groups()

    # Always provide the URI scheme in lowercase characters.
    scheme = scheme.lower()

    # Always provide the host, if any, in lowercase characters.
    host = host.lower()
    if host and host[-1] == '.':
        host = host[:-1]
    # take care about IDN domains
    host = host.decode(charset).encode('idna')  # IDN -> ACE

    # Only perform percent-encoding where it is essential.
    # Always use uppercase A-through-F characters when percent-encoding.
    # All portions of the URI must be utf-8 encoded NFC from Unicode strings
    path = quote(_clean(path), "~:/?#[]@!$&'()*+,;=")
    fragment = quote(_clean(fragment), "~")

    # note care must be taken to only encode & and = characters as values
    query = "&".join(["=".join([quote(_clean(t), "~:/?#[]@!$'()*+,;=") \
                                for t in q.split("=", 1)]) for q in query.split("&")])

    # Prevent dot-segments appearing in non-relative URI paths.
    if scheme in ["", "http", "https", "ftp", "file"]:
        output = []
        for part in path.split('/'):
            if part == "":
                if not output:
                    output.append(part)
            elif part == ".":
                pass
            elif part == "..":
                if len(output) > 1:
                    output.pop()
            else:
                output.append(part)
        if part in ["", ".", ".."]:
            output.append("")
        path = '/'.join(output)

    # For schemes that define a default authority, use an empty authority if
    # the default is desired.
    if userinfo in ["@", ":@"]:
        userinfo = ""

    # For schemes that define an empty path to be equivalent to a path of "/",
    # use "/".
    if path == "" and scheme in ["http", "https", "ftp", "file"]:
        path = "/"

    # For schemes that define a port, use an empty port if the default is
    # desired
    if port and scheme in list(default_port.keys()):
        if port.isdigit():
            port = str(int(port))
            if int(port) == default_port[scheme]:
                port = ''

    # Put it all back together again
    auth = (userinfo or "") + host
    if port:
        auth += ":" + port
    if url.endswith("#") and query == "" and fragment == "":
        path += "#"
    return urlunsplit((scheme, auth, path, query, fragment))

def _parse_hostname(url, include_port=False):
    """ Parses the hostname out of a URL."""
    if url:
        parsed_url = urlparse((url))
        return parsed_url.netloc if include_port else parsed_url.hostname

def _is_http_url(url):
    if url:
        return urlparse(url).scheme in ['http', 'https']

def _unpack(obj_or_seq, key=None, flatten=False):
    """ Turns a list of single item dicts in a list of the dict's values."""

    # The trivial case (passed in None, return None)
    if not obj_or_seq:
        return None

    # We assume it's a sequence
    new_list = []
    for obj in obj_or_seq:
        value = _unpack_obj(obj, key, flatten)
        new_list.extend(value)

    return new_list

def _unpack_obj(obj, key=None, flatten=False):
    try:
        if key:
            value = [obj.get(key)]
        else:
            value = list(obj.values())
    except AttributeError:
        value = [obj]

    # Flatten any lists if directed to do so
    if value and flatten:
        value = [item for sublist in value for item in sublist]

    return value

def _remove_non_ascii(s):
    return ''.join(i for i in s if ord(i) < 128)

def _tostr(obj):
    if not obj:
        return ''
    if isinstance(obj, list):
        return ', '.join(map(_tostr, obj))
    return str(obj)




# This function is a workaround to deal with what's typically described as a
# problem with the web server closing a connection. This is problem
# experienced with www.arcgis.com (first encountered 12/13/2012). The problem
# and workaround is described here:
# http://bobrochel.blogspot.com/2010/11/bad-servers-chunked-encoding-and.html
def _patch_http_response_read(func):
    def inner(*args):
        try:
            return func(*args)
        except http_client.IncompleteRead as e:
            return e.partial

    return inner
http_client.HTTPResponse.read = _patch_http_response_read(http_client.HTTPResponse.read)
